"""
CNN Training Script. 
Improvement from Imagenet script is taken from ross wightman
One of the orignal script is here
https://github.com/rwightman/pytorch-image-models/blob/master/train.py
https://github.com/pytorch/examples/tree/master/imagenet
This is just simplified version of both of them, trying to provide flexibility and modularity.

Some Improvement I am trying to make
1. Support mixed precision training with PyTorch 1.6 (soon).
2. Early Stopping. (Done)
3. Add torchvision support. (Done)
4. fit() function (Done)
5. Keras like history object. (TODO)
6. Sanity Checker. (Done)
7. Gradient Penalty. (Done)
8. Mixed Precision Training. (Done)
"""

import torch
from torch.cuda import amp
from pytorch_cnn_trainer import utils
from tqdm import tqdm
import time

# import model
from collections import OrderedDict

__all__ = [
    "train_step",
    "val_step",
    "fit",
    "train_sanity_fit",
    "val_sanity_fit",
    "sanity_fit",
]


def train_step(
    model,
    train_loader,
    criterion,
    device,
    optimizer,
    scheduler=None,
    num_batches: int = None,
    log_interval: int = 100,
    grad_penalty: bool = False,
    fp16_scaler=None,
):
    """
    Performs one step of training. Calculates loss, forward pass, computes gradient and returns metrics.
    Args:
        model : A pytorch CNN Model.
        train_loader : Train loader.
        criterion : Loss function to be optimized.
        device : "cuda" or "cpu"
        optimizer : Torch optimizer to train.
        scheduler : Learning rate scheduler.
        num_batches : (optional) Integer To limit training to certain number of batches.
        log_interval : (optional) Defualt 100. Integer to Log after specified batch ids in every batch.
        grad_penalty : (optional) To penalize with l2 norm for big gradients.
        fp16_scaler: (optional)  Pass torch.cuda.amp.GradScaler() for fp16 precision Training.
    """

    start_train_step = time.time()

    model.train()
    last_idx = len(train_loader) - 1
    batch_time_m = utils.AverageMeter()
    # data_time_m = utils.AverageMeter()
    losses_m = utils.AverageMeter()
    top1_m = utils.AverageMeter()
    top5_m = utils.AverageMeter()
    cnt = 0
    batch_start = time.time()
    # num_updates = epoch * len(loader)

    for batch_idx, (inputs, target) in enumerate(train_loader):
        last_batch = batch_idx == last_idx
        # data_time_m.update(time.time() - batch_start)
        inputs = inputs.to(device)
        target = target.to(device)

        # zero the parameter gradients
        optimizer.zero_grad()

        if fp16_scaler is not None:
            with amp.autocast():
                output = model(inputs)
                loss = criterion(output, target)
                # Scale the loss using Grad Scaler

            if grad_penalty is True:
                # Scales the loss for autograd.grad's backward pass, resulting in scaled grad_params
                scaled_grad_params = torch.autograd.grad(
                    fp16_scaler.scale(loss), model.parameters(), create_graph=True
                )
                # Creates unscaled grad_params before computing the penalty. scaled_grad_params are
                # not owned by any optimizer, so ordinary division is used instead of fp16_scaler.unscale_:
                inv_scale = 1.0 / fp16_scaler.get_scale()
                grad_params = [p * inv_scale for p in scaled_grad_params]
                # Computes the penalty term and adds it to the loss
                with amp.autocast():
                    grad_norm = 0
                    for grad in grad_params:
                        grad_norm += grad.pow(2).sum()

                    grad_norm = grad_norm.sqrt()
                    loss = loss + grad_norm

            fp16_scaler.scale(loss).backward()
            # Step using fp16_scaler.step()
            fp16_scaler.step(optimizer)
            # Update for next iteration
            fp16_scaler.update()

        else:
            output = model(inputs)
            loss = criterion(output, target)

            if grad_penalty is True:
                # Create gradients
                grad_params = torch.autograd.grad(
                    loss, model.parameters(), create_graph=True
                )
                # Compute the L2 Norm as penalty and add that to loss
                grad_norm = 0
                for grad in grad_params:
                    grad_norm += grad.pow(2).sum()
                grad_norm = grad_norm.sqrt()
                loss = loss + grad_norm

            loss.backward()
            optimizer.step()

        if scheduler is not None:
            scheduler.step()

        cnt += 1
        acc1, acc5 = utils.accuracy(output, target, topk=(1, 5))

        top1_m.update(acc1.item(), output.size(0))
        top5_m.update(acc5.item(), output.size(0))
        losses_m.update(loss.item(), inputs.size(0))

        batch_time_m.update(time.time() - batch_start)
        batch_start = time.time()
        if last_batch or batch_idx % log_interval == 0:  # If we reach the log intervel
            print(
                "Batch Train Time: {batch_time.val:.3f} ({batch_time.avg:.3f})  "
                "Loss: {loss.val:>7.4f} ({loss.avg:>6.4f})  "
                "Top 1 Accuracy: {top1.val:>7.4f} ({top1.avg:>7.4f})  "
                "Top 5 Accuracy: {top5.val:>7.4f} ({top5.avg:>7.4f})".format(
                    batch_time=batch_time_m, loss=losses_m, top1=top1_m, top5=top5_m
                )
            )

        if num_batches is not None:
            if cnt >= num_batches:
                end_train_step = time.time()
                metrics = OrderedDict(
                    [("loss", losses_m.avg), ("top1", top1_m.avg), ("top5", top5_m.avg)]
                )
                print("Done till {} train batches".format(num_batches))
                print(
                    "Time taken for train step = {} sec".format(
                        end_train_step - start_train_step
                    )
                )
                return metrics

    metrics = OrderedDict(
        [("loss", losses_m.avg), ("top1", top1_m.avg), ("top5", top5_m.avg)]
    )
    end_train_step = time.time()
    print(
        "Time taken for train step = {} sec".format(end_train_step - start_train_step)
    )
    return metrics


def val_step(
    model, val_loader, criterion, device, num_batches=None, log_interval: int = 100
):
    """
    Performs one step of validation. Calculates loss, forward pass and returns metrics.
    Args:
        model : A pytorch CNN Model.
        val_loader : Validation loader.
        criterion : Loss function to be optimized.
        device : "cuda" or "cpu"
        num_batches : (optional) Integer To limit validation to certain number of batches.
        log_interval : (optional) Defualt 100. Integer to Log after specified batch ids in every batch.
    """
    start_test_step = time.time()
    last_idx = len(val_loader) - 1
    batch_time_m = utils.AverageMeter()
    # data_time_m = utils.AverageMeter()
    losses_m = utils.AverageMeter()
    top1_m = utils.AverageMeter()
    top5_m = utils.AverageMeter()
    cnt = 0
    model.eval()
    batch_start = time.time()
    with torch.no_grad():
        for batch_idx, (inputs, target) in enumerate(val_loader):
            last_batch = batch_idx == last_idx
            inputs = inputs.to(device)
            target = target.to(device)

            output = model(inputs)
            if isinstance(output, (tuple, list)):
                output = output[0]

            loss = criterion(output, target)
            cnt += 1
            acc1, acc5 = utils.accuracy(output, target, topk=(1, 5))
            reduced_loss = loss.data

            losses_m.update(reduced_loss.item(), inputs.size(0))
            top1_m.update(acc1.item(), output.size(0))
            top5_m.update(acc5.item(), output.size(0))
            batch_time_m.update(time.time() - batch_start)

            batch_start = time.time()

            if (
                last_batch or batch_idx % log_interval == 0
            ):  # If we reach the log intervel
                print(
                    "Batch Inference Time: {batch_time.val:.3f} ({batch_time.avg:.3f})  "
                    "Loss: {loss.val:>7.4f} ({loss.avg:>6.4f})  "
                    "Top 1 Accuracy: {top1.val:>7.4f} ({top1.avg:>7.4f})  "
                    "Top 5 Accuracy: {top5.val:>7.4f} ({top5.avg:>7.4f})".format(
                        batch_time=batch_time_m, loss=losses_m, top1=top1_m, top5=top5_m
                    )
                )

            if num_batches is not None:
                if cnt >= num_batches:
                    end_test_step = time.time()
                    metrics = OrderedDict(
                        [
                            ("loss", losses_m.avg),
                            ("top1", top1_m.avg),
                            ("top5", top5_m.avg),
                        ]
                    )
                    print("Done till {} validation batches".format(num_batches))
                    print(
                        "Time taken for validation step = {} sec".format(
                            end_test_step - start_test_step
                        )
                    )
                    return metrics

        metrics = OrderedDict(
            [("loss", losses_m.avg), ("top1", top1_m.avg), ("top5", top5_m.avg)]
        )
        print("Finished the validation epoch")

    end_test_step = time.time()
    print(
        "Time taken for validation step = {} sec".format(
            end_test_step - start_test_step
        )
    )
    return metrics


def fit(
    epochs,
    model,
    train_loader,
    valid_loader,
    criterion,
    device,
    optimizer,
    scheduler=None,
    early_stopper=None,
    num_batches: int = None,
    log_interval: int = 100,
    grad_penalty: bool = False,
    use_fp16: bool = False,
    swa_start: int = None,
    swa_scheduler=None,
):
    """
    A fit function that performs training for certain number of epochs.
    Args:
        epochs: Number of epochs to train.
        model : A pytorch CNN Model.
        train_loader : Train loader.
        val_loader : Validation loader.
        criterion : Loss function to be optimized.
        device : "cuda" or "cpu"
        optimizer : PyTorch optimizer.
        scheduler : (optional) Learning Rate scheduler.
        early_stopper: (optional) A utils provied early stopper, based on validation loss.
        num_batches : (optional) Integer To limit validation to certain number of batches.
        log_interval : (optional) Defualt 100. Integer to Log after specified batch ids in every batch.
        use_fp16 : (optional) To use Mixed Precision Training using float16 dtype.
        swa_start : (optional) To use Stochastic Weighted Averaging while Training
        swa_scheduler : (optional) A torch.optim.swa_utils.scheduler to be used during SWA Training epochs.
    """

    if swa_start is not None:
        swa_model = torch.optim.swa_utils.AveragedModel(model)
        print("Training with Stochastic Weighted averaging (SWA) Scheduler")

    if use_fp16 is True:
        print("Training with Mixed precision fp16 scaler")
        scaler = amp.GradScaler()
    else:
        scaler = None

    for epoch in tqdm(range(epochs)):
        print()
        print("Training Epoch = {}".format(epoch))
        if swa_start is not None:
            if epoch > swa_start:
                train_metrics = train_step(
                    model,
                    train_loader,
                    criterion,
                    device,
                    optimizer,
                    scheduler=None,
                    num_batches=num_batches,
                    log_interval=log_interval,
                    grad_penalty=grad_penalty,
                    fp16_scaler=scaler,
                )

                swa_model.update_parameters(model)
                swa_scheduler.step()

            else:
                train_metrics = train_step(
                    model,
                    train_loader,
                    criterion,
                    device,
                    optimizer,
                    scheduler,
                    num_batches,
                    log_interval,
                    grad_penalty,
                    fp16_scaler=scaler,
                )

        else:
            train_metrics = train_step(
                model,
                train_loader,
                criterion,
                device,
                optimizer,
                scheduler,
                num_batches,
                log_interval,
                grad_penalty,
                fp16_scaler=scaler,
            )

        print()
        print("Validating Epoch = {}".format(epoch))
        valid_metrics = val_step(
            model, valid_loader, criterion, device, num_batches, log_interval
        )

        validation_loss = valid_metrics["loss"]
        if early_stopper is not None:
            early_stopper(validation_loss, model=model)

            if early_stopper.early_stop:
                print("Saving Model and Early Stopping")
                print("Early Stopping. Ran out of Patience for validation loss")
                break

        print("Done Training, Model Saved to Disk")

    return True  # For now, we should probably return an history object like keras.


def train_sanity_fit(
    model,
    train_loader,
    criterion,
    device,
    num_batches: int = None,
    log_interval: int = 100,
    grad_penalty: bool = False,
    use_fp16: bool = False,
):
    """
    Performs Sanity fit over train loader.
    Use this to dummy check your fit function. It does not calculate metrics, timing, or does checkpointing.
    It iterates over both train_loader and valid_loader for given batches.
    Note: - It does not to loss.backward().
    Args:
        model : A pytorch CNN Model.
        train_loader : Train loader.
        criterion : Loss function to be optimized.
        device : "cuda" or "cpu"
        num_batches : (optional) Integer To limit sanity fit over certain batches. Useful is data is too big even for sanity check.
        log_interval : (optional) Defualt 100. Integer to Log after specified batch ids in every batch.
        use_fp16: : (optional) If True uses PyTorch native mixed precision Training.
    """
    model.train()
    cnt = 0
    last_idx = len(train_loader) - 1
    train_sanity_start = time.time()

    if use_fp16 is True:
        scaler = amp.GradScaler()

    for batch_idx, (inputs, target) in enumerate(train_loader):
        last_batch = batch_idx == last_idx
        # data_time_m.update(time.time() - batch_start)
        inputs = inputs.to(device)
        target = target.to(device)
        output = model(inputs)

        if use_fp16 is True:
            with amp.autocast():
                output = model(inputs)
                loss = criterion(output, target)

        else:
            loss = criterion(output, target)
            if grad_penalty is True:
                # Create gradients
                grad_params = torch.autograd.grad(
                    loss, model.parameters(), create_graph=True
                )
                # Compute the L2 Norm as penalty and add that to loss
                grad_norm = 0
                for grad in grad_params:
                    grad_norm += grad.pow(2).sum()
                grad_norm = grad_norm.sqrt()
                loss = loss + grad_norm

        cnt += 1

        if last_batch or batch_idx % log_interval == 0:
            print(
                "Train Sanity check passed for batch till {} batches".format(batch_idx)
            )

        if num_batches is not None:
            if cnt >= num_batches:
                print("Sanity check passed till {} train batches".format(cnt))
                print("All specicied batches done.")
                train_sanity_end = time.time()
                print(
                    "Training Sanity check passed in time {} !!".format(
                        train_sanity_end - train_sanity_start
                    )
                )
                return True

    train_sanity_end = time.time()
    print(
        "Training Sanity check passed in time {} !!".format(
            train_sanity_end - train_sanity_start
        )
    )
    return True


def val_sanity_fit(
    model,
    valid_loader,
    criterion,
    device,
    num_batches: int = None,
    log_interval: int = 100,
):
    """
    Performs Sanity fit over valid loader.
    Use this to dummy check your fit function. It does not calculate metrics, timing, or does checkpointing.
    It iterates over both train_loader and valid_loader for given batches.
    Note: - It does not to loss.backward().
    Args:
        model : A pytorch CNN Model.
        val_loader : Validation loader.
        criterion : Loss function to be optimized.
        device : "cuda" or "cpu"
        num_batches : (optional) Integer To limit sanity fit over certain batches. Useful is data is too big even for sanity check.
        log_interval : (optional) Defualt 100. Integer to Log after specified batch ids in every batch.
    """
    model.eval()
    cnt = 0
    last_idx = len(valid_loader) - 1
    val_sanity_start = time.time()

    with torch.no_grad():
        for batch_idx, (inputs, target) in enumerate(valid_loader):
            last_batch = batch_idx == last_idx
            # data_time_m.update(time.time() - batch_start)
            inputs = inputs.to(device)
            target = target.to(device)
            output = model(inputs)

            loss = criterion(output, target)
            cnt += 1

            if last_batch or batch_idx % log_interval == 0:
                print(
                    "Validation Sanity check passed for batch till {} batches".format(
                        batch_idx
                    )
                )

            if num_batches is not None:
                if cnt >= num_batches:
                    print("Sanity check passed till {} validation batches".format(cnt))
                    print("All specicied batches done.")
                    val_sanity_end = time.time()
                    print(
                        "Training Sanity check passed in time {} !!".format(
                            val_sanity_end - val_sanity_start
                        )
                    )
                    return True

    val_sanity_end = time.time()
    print(
        "Training Sanity check passed in time {} !!".format(
            val_sanity_end - val_sanity_start
        )
    )
    return True


def sanity_fit(
    model,
    train_loader,
    valid_loader,
    criterion,
    device,
    num_batches: int = None,
    log_interval: int = 100,
    grad_penalty: bool = False,
    use_fp16: bool = False,
):
    """
    Performs Sanity fit over train loader and valid loader.
    Use this to dummy check your fit function. It does not calculate metrics, timing, or does checkpointing.
    It iterates over both train_loader and valid_loader for given batches.
    Note: - It does not to loss.backward().
    Args:
        model : A pytorch CNN Model.
        train_loader : Training loader.
        val_loader : Validation loader.
        criterion : Loss function to be optimized.
        device : "cuda" or "cpu"
        num_batches : (optional) Integer To limit sanity fit over certain batches. Useful is data is too big even for sanity check.
        log_interval : (optional) Defualt 100. Integer to Log after specified batch ids in every batch.
    """
    # Train sanity check
    ts = train_sanity_fit(
        model,
        train_loader,
        criterion,
        device,
        num_batches,
        log_interval,
        grad_penalty,
        use_fp16,
    )
    vs = val_sanity_fit(
        model, valid_loader, criterion, device, num_batches, log_interval
    )
    # This line would run in case both pass otherwise it will create error naturally from torch.
    return True



from pytorch_cnn_trainer.dataset import *
from pytorch_cnn_trainer.engine import *
from pytorch_cnn_trainer.model_factory import *
from pytorch_cnn_trainer.utils import *
from pytorch_cnn_trainer.lightning_trainer import *



# TODO add ImageFolder dataset
# Add code for Mapping using dataframe containaing id and target.

import torchvision
from torchvision import datasets
from torch.utils.data import Dataset
import os
import torch
from PIL import Image

__all__ = [
    "create_fashion_mnist_dataset",
    "create_folder_dataset",
    "create_loaders",
    "CSVDataset",
]


def create_fashion_mnist_dataset(train_transforms, valid_transforms):
    """ Creates Fashion MNIST train dataset and a test dataset. 
    Args: 
    train_transforms: Transforms to be applied to train dataset.
    test_transforms: Transforms to be applied to test dataset.
    """

    # This code can be re-used for other torchvision Image Dataset too.
    train_set = torchvision.datasets.FashionMNIST(
        "./data", download=True, train=True, transform=train_transforms
    )

    valid_set = torchvision.datasets.FashionMNIST(
        "./data", download=True, train=False, transform=valid_transforms
    )

    return train_set, valid_set


def create_cifar10_dataset(train_transforms, valid_transforms):
    """ Creates CIFAR10 train dataset and a test dataset. 
    Args: 
    train_transforms: Transforms to be applied to train dataset.
    test_transforms: Transforms to be applied to test dataset.
    """
    # This code can be re-used for other torchvision Image Dataset too.
    train_set = torchvision.datasets.CIFAR10(
        "./data", download=True, train=True, transform=train_transforms
    )

    valid_set = torchvision.datasets.CIFAR10(
        "./data", download=True, train=False, transform=valid_transforms
    )

    return train_set, valid_set


def create_folder_dataset(root_dir, transforms, split: float = 0.8, **kwargs):
    """
    Creates Train and Validation Dataset from a Root folder 
    Arrange dataset as follows: -
    root/class_a/image01.png
    root/class_b/image01.png
    
    Creates train and validation dataset from this root dir.
    This applies same transforms to both train and validation

    Args:
    root_dir : Root directory of the dataset to read from
    transforms: Transforms to be applied to train and validation datasets.
    split: Float number denoting percentage of train items 

    """
    complete_dataset = datasets.ImageFolder(root_dir, transform=transforms)
    train_split = len(complete_dataset) * split
    valid_split = len(complete_dataset) * (1 - split)

    train_set, valid_set = torch.utils.data.random_split(
        complete_dataset, [train_split, valid_split]
    )

    return train_set, valid_set


def create_loaders(
    train_dataset,
    valid_dataset,
    train_batch_size=32,
    valid_batch_size=32,
    num_workers=1,
    **kwargs
):
    """
    Creates train loader and test loader from train and test datasets
    Args:
    train_dataset: Torchvision train dataset.
    valid_dataset: Torchvision valid dataset.
    train_batch_size (int) : Default 32, Training Batch size
    valid_batch_size (int) : Default 32, Validation Batch size
    num_workers (int) : Defualt 1, Number of workers for training and validation.
    """
    train_loader = torch.utils.data.DataLoader(
        train_dataset, train_batch_size, shuffle=True, num_workers=num_workers
    )
    valid_loader = torch.utils.data.DataLoader(
        valid_dataset, valid_batch_size, shuffle=False, num_workers=num_workers
    )
    return train_loader, valid_loader


class CSVDataset(Dataset):
    def __init__(self, df, data_dir, target, transform):
        super().__init__()
        self.df = df
        self.data_dir = data_dir
        self.transform = transform

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        img_name = self.df.image_id[idx]
        label = self.df.target[idx]

        img_path = os.path.join(self.data_dir, img_name)
        image = Image.open(img_path)
        image = self.transform(image)

        return image, label


# if __name__ == "__main__":
#     train_set, valid_set = create_fashion_mnist_dataset(config.train_transforms, config.valid_transforms)



# We can use any model timm models or torchvision models.
# Model should have AdaptiveAvgPool2d layer and then Linear Layer
# We will remove the Linear Layer and create a new Linear Layer with num_classes

import timm
import torchvision
import torch.nn as nn

__all__ = ["create_timm_model", "create_torchvision_model", "fine_tune_torchvision"]


def create_timm_model(
    model_name: str, num_classes: int, in_channels: int = 3, pretrained: bool = True,
):
    """
    Creates a model from PyTorch Image Models repository.
    To know which models are supported print(timm.list_models())
    Args:
        model_name (str) : Name of the model. E.g. efficientnet_b3
        num_classes (int) : Number of classes for classification.
        in_channels (int) : Defualt 3. Number of channels of images, 1 for grayscale, 3 for RGB
        pretrained (bool) : If true uses modelwweights pretrained on ImageNet.
    """
    model = timm.create_model(
        model_name,
        pretrained=pretrained,
        num_classes=num_classes,
        in_chans=in_channels,
    )
    return model


# Use this when adaptive avg_pooling is not available or you want to pass custom CNN
# This can work with adaptive case too. So it is a bit generic case.
def _create_backbone_generic(model, out_channels: int):
    # Here out_channels is a mandatory argument.
    modules_total = list(model.children())
    modules = modules_total[:-1]
    ft_backbone = nn.Sequential(*modules)
    ft_backbone.out_channels = out_channels
    return ft_backbone


# Use this when you have Adaptive Pooling layer in End.
# When Model.features is not applicable.
def _create_backbone_adaptive(model, out_channels: int = None):
    # Out channels is optional can pass it if user knows it
    # print(list(model.children())[-1].in_features)
    if out_channels is None:
        # Tries to infer the out_channels from layer before adaptive avg pool.
        modules_total = list(model.children())
        out_channels = modules_total[-1].in_features
    return _create_backbone_generic(model, out_channels=out_channels)


# Use this when model.features function is available as in mobilenet and vgg
# Again it is mandatory for out_channels here
def _create_backbone_features(model, out_channels: int):
    ft_backbone = model.features
    ft_backbone.out_channels = out_channels
    return ft_backbone


def _create_torchvision_backbone(
    model_name: str, num_classes: int, pretrained: bool = True
):
    """
    Creates CNN backbone from Torchvision.
    Args:
        model_name (str) : Name of the model. E.g. resnet18
        num_classes (int) : Number of classes for classification.
        pretrained (bool) : If true uses modelwweights pretrained on ImageNet.
    """
    if model_name == "mobilenet":
        mobile_net = torchvision.models.mobilenet_v2(pretrained=pretrained)
        out_channels = 1280
        ft_backbone = _create_backbone_features(mobile_net, 1280)
        return ft_backbone, out_channels

    elif model_name in [
        "vgg11",
        "vgg13",
        "vgg16",
        "vgg19",
    ]:
        out_channels = 512

        if model_name == "vgg11":
            vgg_net = torchvision.models.vgg11(pretrained=pretrained)
        elif model_name == "vgg13":
            vgg_net = torchvision.models.vgg13(pretrained=pretrained)
        elif model_name == "vgg16":
            vgg_net = torchvision.models.vgg16(pretrained=pretrained)
        elif model_name == "vgg19":
            vgg_net = torchvision.models.vgg19(pretrained=pretrained)

        ft_backbone = _create_backbone_features(vgg_net, out_channels)
        return ft_backbone, out_channels

    elif model_name in ["resnet18", "resnet34"]:
        out_channels = 512
        if model_name == "resnet18":
            resnet_net = torchvision.models.resnet18(pretrained=pretrained)
        elif model_name == "resnet34":
            resnet_net = torchvision.models.resnet34(pretrained=pretrained)

        ft_backbone = _create_backbone_adaptive(resnet_net, out_channels)
        return ft_backbone, out_channels

    elif model_name in [
        "resnet50",
        "resnet101",
        "resnet152",
        "resnext50_32x4d",
        "resnext101_32x8d",
    ]:
        out_channels = 2048
        if model_name == "resnet50":
            resnet_net = torchvision.models.resnet50(pretrained=pretrained)
        elif model_name == "resnet101":
            resnet_net = torchvision.models.resnet101(pretrained=pretrained)
        elif model_name == "resnet152":
            resnet_net = torchvision.models.resnet152(pretrained=pretrained)
        elif model_name == "resnext50_32x4d":
            resnet_net = torchvision.models.resnext50_32x4d(pretrained=pretrained)
        elif model_name == "resnext101_32x8d":
            resnet_net = torchvision.models.resnext101_32x8d(pretrained=pretrained)

        ft_backbone = _create_backbone_adaptive(resnet_net, 2048)
        return ft_backbone, out_channels

    elif model_name in ["mnasnet0_5", "mnasnet0_75", "mnasnet1_0", "mnasnet1_3"]:
        out_channels = 1280
        if model_name == "mnasnet0_5":
            mnasnet_net = torchvision.models.mnasnet0_5(pretrained=pretrained)
        elif model_name == "mnasnet0_75":
            mnasnet_net = torchvision.models.mnasnet0_75(pretrained=pretrained)
        elif model_name == "mnasnet1_0":
            mnasnet_net = torchvision.models.mnasnet1_0(pretrained=pretrained)
        elif model_name == "mnasnet1_3":
            mnasnet_net = torchvision.models.mnasnet1_3(pretrained=pretrained)

        ft_backbone = _create_backbone_adaptive(mnasnet_net, 1280)
        return ft_backbone, out_channels

    else:
        raise ValueError("No such model implemented in torchvision")


class fine_tune_torchvision(nn.Module):
    def __init__(self, model_name: str, num_classes: int, pretrained: bool = True):
        super().__init__()
        # old_model = torchvision.models.resnet18(pretrained=True)
        # self.bottom = nn.Sequential(*list(old_model.children())[:-1])
        self.bottom, self.out_channels = _create_torchvision_backbone(
            model_name, num_classes, pretrained=pretrained
        )
        self.top = nn.Linear(self.out_channels, num_classes)

    def forward(self, x):
        x = self.bottom(x)
        x = self.top(x.view(-1, self.out_channels))
        return x


# Wrapper function for consistency with timm models
# We do not pass in_channels here since torchvision models is not supported with it.
def create_torchvision_model(
    model_name: str, num_classes: int, pretrained: bool = True,
):
    """
    Creates CNN model from Torchvision. It replaces the top classification layer with num_classes you need.
    Args:
        model_name (str) : Name of the model. E.g. resnet18
        num_classes (int) : Number of classes for classification.
        pretrained (bool) : If true uses modelwweights pretrained on ImageNet.
    """
    model = fine_tune_torchvision(model_name, num_classes, pretrained)
    return model


# if __name__ == "__main__":
# ft_model = fine_tune_torchvision("mobilenet", 10)
# print(ft_model)
# print("Done")



import numpy as np
import torch
import random
import matplotlib.pyplot as plt
import os

__all__ = [
    "seed_everything",
    "AverageMeter",
    "accuracy",
    "EarlyStopping",
    "matplotlib_imshow",
    "print_size_of_model",
]


def seed_everything(seed):
    """
    Makes code deterministic using a given seed. Internally sets all seeds of torch, numpy and random.
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)
    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    return [correct[:k].view(-1).float().sum(0) * 100.0 / batch_size for k in topk]


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print("\t".join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = "{:" + str(num_digits) + "d}"
        return "[" + fmt + "/" + fmt.format(num_batches) + "]"


class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""

    def __init__(self, patience=7, verbose=False, delta=0.0001, path="checkpoint.pt"):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement. 
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.path = path

    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
        elif score < self.best_score + self.delta:
            self.counter += 1
            print(f"EarlyStopping counter: {self.counter} out of {self.patience}")
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss, model):
        """Saves model when validation loss decrease."""
        if self.verbose:
            print(
                f"Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ..."
            )
        torch.save(model.state_dict(), self.path)
        self.val_loss_min = val_loss


def print_size_of_model(model):
    torch.save(model.state_dict(), "temp.p")
    print("Size (MB):", os.path.getsize("temp.p") / 1e6)
    os.remove("temp.p")


def matplotlib_imshow(img, one_channel=False):
    if one_channel:
        img = img.mean(dim=0)
    img = img / 2 + 0.5  # unnormalize
    npimg = img.numpy()
    if one_channel:
        plt.imshow(npimg, cmap="Greys")
    else:
        plt.imshow(np.transpose(npimg, (1, 2, 0)))
